import contextlib
import io
import os
import posixpath
import stat
import threading
import time
from collections.abc import Iterable
from typing import IO, Any, Callable, Optional, Union, cast, final

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.StenciledFile import RawStenciledFile, StenciledFile
from ratarmountcore.utils import get_groupid, get_userid, overrides


def _get_mode(fileobj: IO[bytes]) -> int:
    return 0o111 | (0o222 if fileobj.writable() else 0) | (0o444 if fileobj.readable() else 0)


@final
class SingleFileMountSource(MountSource):
    """MountSource exposing a single file as a mount source."""

    # It makes no sense to merge this class into SubvolumesMountSource because of the file locking and file size
    # for each file that needs to be saved and it would require adding if-else into almost every method in
    # SubvolumesMountSource. Having a simple SingleFileMountSource to be used with SubvolumesMountSource is far
    # more elegant. The only difference might be how to handle multiple mount points/files per parent folder.
    # This is currently not possible because SingleFileMountSource implies a single folder containing a single file.
    # It could be emulated with UnionMountSource or by extending SingleFileMountSource to support multiple files,
    # but then it would add inconsistencies when these multiple files have different full paths / implied parents...

    def __init__(self, path: str, fileobj: Union[IO[bytes], Callable[[int], IO[bytes]]]):
        """
        fileobj: The given file object to be mounted. It may be advisable for this file object to be unbuffered
                 because opening file objects via this mount source will add additional buffering if not disabled.
        """
        self.path = posixpath.normpath('/' + path).lstrip('/')
        if not self.path or '/' in self.path:
            raise ValueError("File object must belong to a non-folder path!")

        self.mtime = time.time()
        self._file_info: Optional[FileInfo] = None
        self._file_object: Optional[IO[bytes]] = None
        self._statfs = {}

        if callable(fileobj):
            self._open_file = fileobj
            return

        # Construct an file-opener callable from an existing file object.

        self._file_info = self._create_file_info(
            isdir=False, size=fileobj.seek(0, io.SEEK_END), mode=_get_mode(fileobj)
        )
        self._file_lock = threading.Lock()
        self._file_object = fileobj

        def open_shared_file(buffering: int):
            if not self._file_object:
                raise RuntimeError("No file object to open!")
            if self._file_info is None:
                raise RuntimeError("File size should have been initialized!")

            # Use StenciledFile so that the returned file objects can be independently seeked!
            if buffering == 0:
                return cast(
                    IO[bytes],
                    RawStenciledFile(
                        fileStencils=[(self._file_object, 0, self._file_info.size)], fileObjectLock=self._file_lock
                    ),
                )
            return cast(
                IO[bytes],
                StenciledFile(
                    fileStencils=[(self._file_object, 0, self._file_info.size)],
                    fileObjectLock=self._file_lock,
                    bufferSize=io.DEFAULT_BUFFER_SIZE if buffering <= 0 else buffering,
                ),
            )

        self._open_file = open_shared_file

        fileno = None
        with contextlib.suppress(Exception):
            fileno = fileobj.fileno()

        if fileno is not None:
            statfs = os.fstat(fileno)
            self._statfs = {
                'f_bsize': statfs.st_blksize,
                'f_frsize': statfs.st_blksize,
                'f_blocks': statfs.st_blocks,
                'f_bfree': 0,
                'f_bavail': 0,
                'f_ffree': 0,
                'f_favail': 0,
            }

    def _create_file_info(self, isdir: bool, size: Optional[int] = None, mode: Optional[int] = None):
        if self._file_info:
            if size is None:
                size = self._file_info.size
            if mode is None:
                mode = self._file_info.mode & 0o777

        if size is None or mode is None:
            if hasattr(self, '_open_file'):
                with self._open_file(-1) as file:
                    if size is None:
                        size = file.seek(0, io.SEEK_END) if file.seekable() else 0
                        if size is None:
                            size = file.tell()
                        if size is None:
                            size = 0
                    if mode is None:
                        mode = _get_mode(file)
            else:
                if size is None:
                    size = 0
                if mode is None:
                    mode = 0o777

        # fmt: off
        return FileInfo(
            size     = size,
            mtime    = self.mtime,
            mode     = mode | (stat.S_IFDIR if isdir else stat.S_IFREG),
            linkname = '',
            uid      = self._file_info.uid if self._file_info else get_userid(),
            gid      = self._file_info.gid if self._file_info else get_groupid(),
            userdata = [],
        )
        # fmt: on

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], dict[str, FileInfo]]]:
        return [self.path] if not path or path == '/' else None

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        if not path or path == '/':
            return self._create_file_info(size=0, isdir=True)
        if path.strip('/') == self.path:
            return self._create_file_info(isdir=False)
        return None

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        if fileInfo != self._create_file_info(size=fileInfo.size, mode=fileInfo.mode & 0o777, isdir=False):
            raise ValueError("Only files may be opened!")
        return self._open_file(buffering)

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def statfs(self) -> dict[str, Any]:
        return self._statfs.copy()

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        if self._file_object:
            self._file_object.close()

    def join_threads(self):
        if hasattr(self._file_object, 'join_threads'):
            self._file_object.join_threads()



# -*- coding: utf-8 -*-

"""Custom Exceptions for PyFAT."""


class PyFATException(Exception):
    """Generic PyFAT Exceptions."""

    def __init__(self, msg: str, errno=None):
        """Construct base class for PyFAT exceptions.

        :param msg: Exception message describing what happened
        :param errno: Error number, mostly based on POSIX errno where feasible
        """
        Exception.__init__(self, msg)
        self.errno = errno


class NotAnLFNEntryException(PyFATException):
    """Indicates that given dir entry cannot be interpreted as LFN entry."""


class BrokenLFNEntryException(PyFATException):
    """Indicates that given LFN entry is invalid."""


class NotAFatEntryException(NotADirectoryError):
    """Custom handling for FAT `NotADirectoryError`'s."""

    def __init__(self, msg: str, free_type: int):
        """Construct base class for PyFAT exceptions.

        :param msg: Exception message describing what happened
        :param free_type: `FATDirectoryEntry._FREE_DIR_ENTRY_MARK` or
                          `FATDirectoryEntry._LAST_DIR_ENTRY_MARK`
        """
        NotADirectoryError.__init__(self, msg)
        self.free_type = free_type



import contextlib
import datetime
import json
import logging
import re
import stat
import sys
import tarfile
import zipfile
from pathlib import Path
from timeit import default_timer as timer
from typing import IO, Any, Optional, Union

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.utils import overrides

try:
    # Importing this patches the zipfile module as a "side" effect!
    import fast_zip_decryption  # pylint: disable=unused-import  # noqa: F401
except (ImportError, Exception):
    with contextlib.suppress(ImportError, Exception):
        import fastzipfile  # pylint: disable=unused-import  # noqa: F401


logger = logging.getLogger(__name__)


class ZipMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes], Path],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[list[str]]       = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        indexMinimumFileCount  : int                       = 1000,
        transform              : Optional[tuple[str, str]] = None,
        **options
    ) -> None:
        # fmt: on
        if 'zipfile' not in sys.modules:
            raise RuntimeError("Did not find the zipfile module. Please use Python 3.7+.")

        if isinstance(fileOrPath, Path):
            fileOrPath = str(fileOrPath)

        # fmt: off
        self.fileObject             = zipfile.ZipFile(fileOrPath, 'r')
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.options                = options
        self.transformPattern       = transform
        # fmt: on

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        ZipMountSource._find_password(self.fileObject, options.get("passwords", []))
        self.files = {info.header_offset: info for info in self.fileObject.infolist()}

        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='ZipMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        isFileObject = not isinstance(fileOrPath, str)

        if self.index.index_is_loaded():
            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding', 'transformPattern']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)

    def _convert_to_row(self, info: "zipfile.ZipInfo") -> tuple:
        mtime = datetime.datetime(*info.date_time, tzinfo=datetime.timezone.utc).timestamp() if info.date_time else 0

        # According to section 4.5.7 in the .ZIP file format specification, links are supported:
        # https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT
        # The Python zipfile module has no API for links: https://bugs.python.org/issue45286
        # However, the file mode exposes whether it's a link and the file mode is shown by ZipInfo.__repr__.
        # For that, it uses the OS-dependent external_attr member. See also the ZIP specification on that:
        # > 4.4.15 external file attributes: (4 bytes)
        # >   The mapping of the external attributes is host-system dependent (see 'version made by').
        # >   For MS-DOS, the low order byte is the MS-DOS directory attribute byte.
        # >   If input came from standard input, this field is set to zero.

        # file_redir is (type, flags, target) or None. Only tested for type == RAR5_XREDIR_UNIX_SYMLINK.
        linkname = ""
        mode = (info.external_attr >> 16) & 0o777
        if mode == 0:
            mode = 0o770 if info.is_dir() else 0o660
        if stat.S_ISLNK(info.external_attr >> 16):
            linkname = self.fileObject.read(info).decode()
            mode = mode | stat.S_IFLNK
        else:
            mode = mode | (stat.S_IFDIR if info.is_dir() else stat.S_IFREG)

        path, name = SQLiteIndex.normpath(self.transform(info.filename)).rsplit("/", 1)

        # Currently, this is unused. The index only is used for getting metadata. (The data offset
        # is already determined and written out in order to possibly speed up reading of encrypted
        # files by implementing the decryption ourselves.)
        # The data offset is deprecated again! Collecting it can add a huge overhead for large zip files
        # because we have to seek to every position and read a few bytes from it. Furthermore, it is useless
        # by itself anyway. We don't even store yet how the data is compressed or encrypted, so we would
        # have to read the local header again anyway!
        dataOffset = 0

        # fmt: off
        fileInfo : tuple = (
            path              ,  # 0  : path
            name              ,  # 1  : file name
            info.header_offset,  # 2  : header offset
            dataOffset        ,  # 3  : data offset
            info.file_size    ,  # 4  : file size
            mtime             ,  # 5  : modification time
            mode              ,  # 6  : file mode / permissions
            0                 ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            linkname          ,  # 8  : linkname
            0                 ,  # 9  : user ID
            0                 ,  # 10 : group ID
            False             ,  # 11 : is TAR (unused?)
            False             ,  # 12 : is sparse
            False             ,  # 13 : is generated (parent folder)
            0                 ,  # 14 : recursion depth
        )
        # fmt: on

        return fileInfo

    def _create_index(self) -> None:
        if logger.isEnabledFor(logging.WARNING):
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()
        self.index.set_file_infos([self._convert_to_row(info) for info in self.fileObject.infolist()])
        self.index.finalize()

        if logger.isEnabledFor(logging.WARNING):
            print(f"Creating offset dictionary for {self.archiveFilePath} took {timer() - t0:.2f}s")

    @staticmethod
    def _find_password(fileobj: "zipfile.ZipFile", passwords):
        # If headers are encrypted, then infolist will simply return an empty list!
        files = fileobj.infolist()
        if not files:
            for password in passwords:
                fileobj.setpassword(password)
                files = fileobj.infolist()
                if files:
                    return password

        # If headers are not encrypted, then try out passwords by trying to open the first file.
        files = [file for file in files if not file.is_dir() and file.file_size > 0]
        if not files:
            return None

        for password in [None, *passwords]:
            fileobj.setpassword(password)
            try:
                with fileobj.open(files[0]) as file:
                    file.read(1)
                return password
            except Exception:
                pass

        raise RuntimeError("Could not find a matching password!")

    @overrides(SQLiteIndexMountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        super().__exit__(exception_type, exception_value, exception_traceback)
        self.fileObject.close()

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        # I do not see any obvious option to zipfile.ZipFile to apply the specified buffer size.
        assert fileInfo.userdata
        extendedFileInfo = fileInfo.userdata[-1]
        assert isinstance(extendedFileInfo, SQLiteIndexedTarUserData)
        info = self.files[extendedFileInfo.offsetheader]
        assert isinstance(info, zipfile.ZipInfo)
        # CPython's zipfile module does handle multiple file objects being opened and reading from the
        # same underlying file object concurrently by using a _SharedFile class that even includes a lock.
        # Very nice!
        # https://github.com/python/cpython/blob/a87c46eab3c306b1c5b8a072b7b30ac2c50651c0/Lib/zipfile/__init__.py#L1569
        return self.fileObject.open(info, 'r')  # https://github.com/pauldmccarthy/indexed_gzip/issues/85

    def _check_metadata(self, metadata: dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(
                json.loads(metadata['arguments']), self, argumentsToCheck=['encoding', 'transformPattern']
            )

        if 'backendName' not in metadata:
            self.index.try_to_open_first_file(lambda path: self.open(self.lookup(path)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""FAT and BPB parsing for files."""

import datetime
import errno

import math
import struct
import threading
import time
import warnings

from contextlib import contextmanager
from io import FileIO, open, BytesIO, IOBase, SEEK_END
from os import PathLike
from typing import Union

from pyfatfs import FAT_OEM_ENCODING, _init_check
from pyfatfs.EightDotThree import EightDotThree
from pyfatfs.FATDirectoryEntry import FATDirectoryEntry, FATLongDirectoryEntry
from pyfatfs.FSInfo import FSInfo
from pyfatfs._exceptions import PyFATException, NotAFatEntryException
from pyfatfs.BootSectorHeader import BootSectorHeader, FAT12BootSectorHeader, \
                                     FAT32BootSectorHeader


class PyFat(object):
    """PyFAT base class, parses generic filesystem information."""

    #: Used as fat_type if unable to detect FAT type
    FAT_TYPE_UNKNOWN = 0
    #: Used as fat_type if FAT12 fs has been detected
    FAT_TYPE_FAT12 = 12
    #: Used as fat_type if FAT16 fs has been detected
    FAT_TYPE_FAT16 = 16
    #: Used as fat_type if FAT32 fs has been detected
    FAT_TYPE_FAT32 = 32

    #: Maps fat_type to BS_FilSysType from FS header information
    FS_TYPES = {FAT_TYPE_UNKNOWN: b"FAT     ",
                FAT_TYPE_FAT12: b"FAT12   ",
                FAT_TYPE_FAT16: b"FAT16   ",
                FAT_TYPE_FAT32: b"FAT32   "}

    #: Possible cluster values for FAT12 partitions
    FAT12_CLUSTER_VALUES = {'FREE_CLUSTER': 0x000,
                            'MIN_DATA_CLUSTER': 0x002,
                            'MAX_DATA_CLUSTER': 0xFEF,
                            'BAD_CLUSTER': 0xFF7,
                            'END_OF_CLUSTER_MIN': 0xFF8,
                            'END_OF_CLUSTER_MAX': 0xFFF}
    FAT12_SPECIAL_EOC = 0xFF0
    #: Possible cluster values for FAT16 partitions
    FAT16_CLUSTER_VALUES = {'FREE_CLUSTER': 0x0000,
                            'MIN_DATA_CLUSTER': 0x0002,
                            'MAX_DATA_CLUSTER': 0xFFEF,
                            'BAD_CLUSTER': 0xFFF7,
                            'END_OF_CLUSTER_MIN': 0xFFF8,
                            'END_OF_CLUSTER_MAX': 0xFFFF}
    #: Possible cluster values for FAT32 partitions
    FAT32_CLUSTER_VALUES = {'FREE_CLUSTER': 0x0000000,
                            'MIN_DATA_CLUSTER': 0x0000002,
                            'MAX_DATA_CLUSTER': 0x0FFFFFEF,
                            'BAD_CLUSTER': 0xFFFFFF7,
                            'END_OF_CLUSTER_MIN': 0xFFFFFF8,
                            'END_OF_CLUSTER_MAX': 0xFFFFFFF}
    #: Maps fat_type to possible cluster values
    FAT_CLUSTER_VALUES = {FAT_TYPE_FAT12: FAT12_CLUSTER_VALUES,
                          FAT_TYPE_FAT16: FAT16_CLUSTER_VALUES,
                          FAT_TYPE_FAT32: FAT32_CLUSTER_VALUES}

    #: FAT16 bit mask for clean shutdown bit
    FAT16_CLEAN_SHUTDOWN_BIT_MASK = 0x8000
    #: FAT16 bit mask for volume error bit
    FAT16_DRIVE_ERROR_BIT_MASK = 0x4000
    #: FAT32 bit mask for clean shutdown bit
    FAT32_CLEAN_SHUTDOWN_BIT_MASK = 0x8000000
    #: FAT32 bit mask for volume error bit
    FAT32_DRIVE_ERROR_BIT_MASK = 0x4000000

    #: Dirty bit in FAT header
    FAT_DIRTY_BIT_MASK = 0x01

    def __init__(self,
                 encoding: str = 'ibm437',
                 offset: int = 0,
                 lazy_load: bool = True):
        """Set up PyFat class instance.

        :param encoding: Define encoding to use for filenames
        :param offset: Offset of the FAT partition in the given file
        :type encoding: str
        :type offset: int
        """
        self.__fp: FileIO = None
        self.__fp_offset = offset
        self._fat_size = 0
        self.bpb_header: BootSectorHeader = None
        self.root_dir = None
        self.root_dir_sector = 0
        self.root_dir_sectors = 0
        self.bytes_per_cluster = 0
        self.first_data_sector = 0
        self.first_free_cluster = 0
        self.fat_type = self.FAT_TYPE_UNKNOWN
        self.fat = {}
        self.initialized = False
        self.encoding = encoding
        self.is_read_only = True
        self.lazy_load = lazy_load
        self.__lock = threading.Lock()

    def __set_fp(self, fp: Union[IOBase, BytesIO]):
        if self.__fp is not None:
            raise PyFATException("Cannot overwrite existing file handle, "
                                 "create new class instance of PyFAT.",
                                 errno=errno.EMFILE)
        self.__fp = fp

    def __seek(self, address: int):
        """Seek to given address with offset."""
        if self.__fp is None:
            raise PyFATException("Cannot seek without a file handle!",
                                 errno=errno.ENXIO)
        self.__fp.seek(address + self.__fp_offset)

    @_init_check
    def read_cluster_contents(self, cluster: int) -> bytes:
        """Read contents of given cluster.

        :param cluster: Cluster number to read contents from
        :returns: Contents of cluster as `bytes`
        """
        sz = self.bytes_per_cluster
        cluster_address = self.get_data_cluster_address(cluster)
        with self.__lock:
            self.__seek(cluster_address)
            return self.__fp.read(sz)

    def __get_clean_shutdown_bitmask(self):
        """Get clean shutdown bitmask for current FS.

        :raises: AttributeError
        """
        return getattr(self, f"FAT{self.fat_type}_CLEAN_SHUTDOWN_BIT_MASK")

    def _is_dirty(self) -> bool:
        """Check whether or not the partition currently is dirty."""
        try:
            clean_shutdown_bitmask = self.__get_clean_shutdown_bitmask()
        except AttributeError:
            # Bit not set on FAT12
            dos_dirty = False
        else:
            dos_dirty = (self.fat[1] &
                         clean_shutdown_bitmask) != clean_shutdown_bitmask

        nt_dirty = (self.bpb_header["BS_Reserved1"] &
                    self.FAT_DIRTY_BIT_MASK) == self.FAT_DIRTY_BIT_MASK

        return dos_dirty or nt_dirty

    def set_fp(self, fp: Union[BytesIO, IOBase]):
        """Open a filesystem from a valid file pointer.

        This allows using in-memory filesystems (e.g., BytesIO).

        :param fp: `FileIO`: Valid `FileIO` object
        """
        if not fp.readable():
            raise PyFATException("Cannot read data from file pointer.",
                                 errno=errno.EACCES)

        if not fp.seekable():
            raise PyFATException("Cannot seek file object.",
                                 errno=errno.EINVAL)

        self.is_read_only = not fp.writable()

        self.__set_fp(fp)

        # Parse BPB & FAT headers of given file
        self.parse_header()

        # Parse FAT
        self._parse_fat()

        # Check for clean shutdown
        if self._is_dirty():
            warnings.warn("Filesystem was not cleanly unmounted on last "
                          "access. Check for data corruption.")

        # Parse root directory
        self.parse_root_dir()

    def open(self, filename: Union[str, PathLike], read_only: bool = False):
        """Open filesystem for usage with PyFat.

        :param filename: `str`: Name of file to open for usage with PyFat.
        :param read_only: `bool`: Force read-only mode of filesystem.
        """
        self.is_read_only = read_only
        if read_only is True:
            mode = 'rb'
        else:
            mode = 'rb+'

        try:
            return self.set_fp(open(filename, mode=mode))
        except OSError as ex:
            raise PyFATException(f"Cannot open given file \'{filename}\'.",
                                 errno=ex.errno)

    @_init_check
    def get_fs_location(self):
        """Retrieve path of opened filesystem."""
        return self.__fp.name

    def _get_total_sectors(self):
        """Get total number of sectors for all FAT sizes."""
        if self.bpb_header["BPB_TotSec16"] != 0:
            return self.bpb_header["BPB_TotSec16"]

        return self.bpb_header["BPB_TotSec32"]

    def _get_fat_size_count(self):
        """Get BPB_FATsz value."""
        if self.bpb_header["BPB_FATSz16"] != 0:
            return self.bpb_header["BPB_FATSz16"]

        try:
            return self.bpb_header["BPB_FATSz32"]
        except KeyError:
            raise PyFATException("Invalid FAT size of 0 detected in header, "
                                 "cannot continue")

    @_init_check
    def _parse_fat(self):
        """Parse information in FAT."""
        # Read all FATs
        fat_size = self.bpb_header["BPB_BytsPerSec"]
        fat_size *= self._fat_size

        # Seek FAT entries
        first_fat_bytes = self.bpb_header["BPB_RsvdSecCnt"]
        first_fat_bytes *= self.bpb_header["BPB_BytsPerSec"]
        fats = []
        for i in range(self.bpb_header["BPB_NumFATs"]):
            with self.__lock:
                self.__seek(first_fat_bytes + (i * fat_size))
                fats += [self.__fp.read(fat_size)]

        if len(fats) < 1:
            raise PyFATException("Invalid number of FATs configured, "
                                 "cannot continue")
        elif len(set(fats)) > 1:
            warnings.warn("One or more FATs differ, filesystem most "
                          "likely corrupted. Using first FAT.")

        # Parse first FAT
        self.bytes_per_cluster = self.bpb_header["BPB_BytsPerSec"] * \
            self.bpb_header["BPB_SecPerClus"]

        if len(fats[0]) != self.bpb_header["BPB_BytsPerSec"] * self._fat_size:
            raise PyFATException("Invalid length of FAT")

        # FAT12: 12 bits (1.5 bytes) per FAT entry
        # FAT16: 16 bits (2 bytes) per FAT entry
        # FAT32: 32 bits (4 bytes) per FAT entry
        fat_entry_size = self.fat_type / 8
        total_entries = int(fat_size // fat_entry_size)
        self.fat = [None] * total_entries

        curr = 0
        cluster = 0
        incr = self.fat_type / 8
        while curr < fat_size:
            offset = curr + incr

            if self.fat_type == self.FAT_TYPE_FAT12:
                fat_nibble = fats[0][int(curr):math.ceil(offset)]
                fat_nibble = fat_nibble.ljust(2, b"\0")
                try:
                    self.fat[cluster] = struct.unpack("<H", fat_nibble)[0]
                except IndexError:
                    # Out of bounds, FAT size is not cleanly divisible by 3
                    # Do not touch last clusters
                    break

                if cluster % 2 == 0:
                    # Even: Keep low 12-bits of word
                    self.fat[cluster] &= 0x0FFF
                else:
                    # Odd: Keep high 12-bits of word
                    self.fat[cluster] >>= 4

                if math.ceil(offset) == (fat_size - 1):
                    # Sector boundary case for FAT12
                    del self.fat[-1]
                    break

            elif self.fat_type == self.FAT_TYPE_FAT16:
                self.fat[cluster] = struct.unpack("<H",
                                                  fats[0][int(curr):
                                                          int(offset)])[0]
            elif self.fat_type == self.FAT_TYPE_FAT32:
                self.fat[cluster] = struct.unpack("<L",
                                                  fats[0][int(curr):
                                                          int(offset)])[0]
                # Ignore first four bits, FAT32 clusters are
                # actually just 28bits long
                self.fat[cluster] &= 0x0FFFFFFF
            else:
                raise PyFATException("Unknown FAT type, cannot continue")

            curr += incr
            cluster += 1

        if None in self.fat:
            raise AssertionError("Unknown error during FAT parsing, please "
                                 "report this error.")

    @_init_check
    def __bytes__(self):
        """Represent current state of FAT as bytes.

        :returns: `bytes` representation of FAT.
        """
        b = b''
        if self.fat_type == self.FAT_TYPE_FAT12:
            for i, e in enumerate(self.fat):
                if i % 2 == 0:
                    b += struct.pack("<H", e)
                else:
                    nibble = b[-1:]
                    nibble = struct.unpack("<B", nibble)[0]
                    b = b[:-1]
                    b += struct.pack("<BB", ((e & 0xF) << 4) | nibble, e >> 4)

        else:
            if self.fat_type == self.FAT_TYPE_FAT16:
                fmt = "H"
            else:
                # FAT32
                fmt = "L"

            b = struct.pack(f"<{fmt * len(self.fat)}",
                            *self.fat)
        return b

    def calc_num_clusters(self, size: int = 0) -> int:
        """Calculate the number of required clusters.

        :param size: `int`: required bytes to allocate
        :returns: Number of required clusters
        """
        num_clusters = size / self.bytes_per_cluster
        num_clusters = math.ceil(num_clusters)

        return num_clusters

    def _fat12_parse_root_dir(self):
        """Parse FAT12/16 root dir entries.

        FAT12/16 has a fixed location of root directory entries
        and is therefore size limited (BPB_RootEntCnt).
        """
        root_dir_byte = self.root_dir_sector * \
            self.bpb_header["BPB_BytsPerSec"]
        self.root_dir.set_cluster(self.root_dir_sector //
                                  self.bpb_header["BPB_SecPerClus"])
        max_bytes = self.bpb_header["BPB_RootEntCnt"] * \
            FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        # Parse all directory entries in root directory
        subdirs, _ = self.parse_dir_entries_in_address(root_dir_byte,
                                                       root_dir_byte +
                                                       max_bytes)
        for dir_entry in subdirs:
            self.root_dir.add_subdirectory(dir_entry)

    def _fat32_parse_root_dir(self):
        """Parse FAT32 root dir entries.

        FAT32 actually has its root directory entries distributed
        across a cluster chain that we need to follow
        """
        root_cluster = self.bpb_header["BPB_RootClus"]
        self.root_dir.set_cluster(root_cluster)

        # Follow root directory cluster chain
        for dir_entry in self.parse_dir_entries_in_cluster_chain(root_cluster):
            self.root_dir.add_subdirectory(dir_entry,
                                           recursive=not self.lazy_load)

    def parse_root_dir(self):
        """Parse root directory entry."""
        root_dir_sfn = EightDotThree()
        root_dir_sfn.set_str_name("")
        dir_attr = FATDirectoryEntry.ATTR_DIRECTORY
        self.root_dir = FATDirectoryEntry(fs=self,
                                          DIR_Name=root_dir_sfn,
                                          DIR_Attr=dir_attr,
                                          DIR_NTRes=0,
                                          DIR_CrtTimeTenth=0,
                                          DIR_CrtTime=0,
                                          DIR_CrtDate=0,
                                          DIR_LstAccessDate=0,
                                          DIR_FstClusHI=0,
                                          DIR_WrtTime=0,
                                          DIR_WrtDate=0,
                                          DIR_FstClusLO=0,
                                          DIR_FileSize=0,
                                          encoding=self.encoding)

        if self.fat_type in [self.FAT_TYPE_FAT12, self.FAT_TYPE_FAT16]:
            self._fat12_parse_root_dir()
        else:
            self._fat32_parse_root_dir()

    def parse_lfn_entry(self,
                        lfn_entry: FATLongDirectoryEntry = None,
                        address: int = 0):
        """Parse LFN entry at given address."""
        dir_hdr_sz = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        with self.__lock:
            self.__seek(address)
            lfn_dir_data = self.__fp.read(dir_hdr_sz)

        lfn_hdr_layout = FATLongDirectoryEntry.FAT_LONG_DIRECTORY_LAYOUT
        lfn_dir_hdr = struct.unpack(lfn_hdr_layout, lfn_dir_data)
        lfn_dir_hdr = dict(zip(FATLongDirectoryEntry.FAT_LONG_DIRECTORY_VARS,
                               lfn_dir_hdr))

        lfn_entry.add_lfn_entry(**lfn_dir_hdr)

    def __parse_dir_entry(self, address):
        """Parse directory entry at given address."""
        with self.__lock:
            self.__seek(address)
            dir_hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE
            dir_data = self.__fp.read(dir_hdr_size)

        dir_hdr = struct.unpack(FATDirectoryEntry.FAT_DIRECTORY_LAYOUT,
                                dir_data)
        dir_hdr = dict(zip(FATDirectoryEntry.FAT_DIRECTORY_VARS, dir_hdr))
        return dir_hdr

    def parse_dir_entries_in_address(self,
                                     address: int = 0,
                                     max_address: int = 0,
                                     tmp_lfn_entry: FATLongDirectoryEntry =
                                     None):
        """Parse directory entries in address range."""
        if tmp_lfn_entry is None:
            tmp_lfn_entry = FATLongDirectoryEntry()

        dir_hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        if max_address == 0:
            max_address = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE

        dir_entries = []

        for hdr_addr in range(address, max_address, dir_hdr_size):
            # Parse each entry
            dir_hdr = self.__parse_dir_entry(hdr_addr)
            dir_sn = EightDotThree(encoding=self.encoding)
            dir_first_byte = dir_hdr["DIR_Name"][0]
            try:
                dir_sn.set_byte_name(dir_hdr["DIR_Name"])
            except NotAFatEntryException as ex:
                # Not a directory of any kind, invalidate temporary LFN entries
                tmp_lfn_entry = FATLongDirectoryEntry()
                if ex.free_type == FATDirectoryEntry.FREE_DIR_ENTRY_MARK:
                    # Empty directory entry,
                    continue
                elif ex.free_type == FATDirectoryEntry.LAST_DIR_ENTRY_MARK:
                    # Last directory entry, do not parse any further
                    break
            else:
                dir_hdr["DIR_Name"] = dir_sn

            # Long File Names
            if FATLongDirectoryEntry.is_lfn_entry(dir_first_byte,
                                                  dir_hdr["DIR_Attr"]):
                self.parse_lfn_entry(tmp_lfn_entry, hdr_addr)
                continue

            # Normal directory entries
            if not tmp_lfn_entry.is_lfn_entry_complete():
                # Ignore incomplete LFN entries altogether
                tmp_lfn_entry = None

            dir_entry = FATDirectoryEntry(fs=self,
                                          encoding=self.encoding,
                                          lfn_entry=tmp_lfn_entry,
                                          lazy_load=self.lazy_load,
                                          **dir_hdr)
            dir_entries += [dir_entry]

            if not self.lazy_load:
                if dir_entry.is_directory() and not dir_entry.is_special():
                    # Iterate all subdirectories except for dot and dotdot
                    cluster = dir_entry.get_cluster()
                    subdirs = self.parse_dir_entries_in_cluster_chain(cluster)
                    for d in subdirs:
                        dir_entry.add_subdirectory(d)

            # Reset temporary LFN entry
            tmp_lfn_entry = FATLongDirectoryEntry()

        return dir_entries, tmp_lfn_entry

    def parse_dir_entries_in_cluster_chain(self, cluster) -> list:
        """Parse directory entries while following given cluster chain."""
        dir_entries = []
        tmp_lfn_entry = FATLongDirectoryEntry()
        max_bytes = (self.bpb_header["BPB_SecPerClus"] *
                     self.bpb_header["BPB_BytsPerSec"])
        for c in self.get_cluster_chain(cluster):
            # Parse all directory entries in chain
            b = self.get_data_cluster_address(c)
            ret = self.parse_dir_entries_in_address(b, b+max_bytes,
                                                    tmp_lfn_entry)
            tmp_dir_entries, tmp_lfn_entry = ret
            dir_entries += tmp_dir_entries

        return dir_entries

    def get_data_cluster_address(self, cluster: int) -> int:
        """Get offset of given cluster in bytes.

        :param cluster: Cluster number as `int`
        :returns: Bytes address location of cluster
        """
        # First two cluster entries are reserved
        sector = (cluster - 2) * self.bpb_header["BPB_SecPerClus"] + \
            self.first_data_sector
        return sector * self.bpb_header["BPB_BytsPerSec"]

    @_init_check
    def get_cluster_chain(self, first_cluster):
        """Follow a cluster chain beginning with the first cluster address."""
        cluster_vals = self.FAT_CLUSTER_VALUES[self.fat_type]
        min_data_cluster = cluster_vals["MIN_DATA_CLUSTER"]
        max_data_cluster = cluster_vals["MAX_DATA_CLUSTER"]
        eoc_min = cluster_vals["END_OF_CLUSTER_MIN"]
        eoc_max = cluster_vals["END_OF_CLUSTER_MAX"]

        i = first_cluster
        while i <= len(self.fat):
            if min_data_cluster <= self.fat[i] <= max_data_cluster:
                # Normal data cluster, follow chain
                yield i
            elif self.fat_type == self.FAT_TYPE_FAT12 and \
                    self.fat[i] == self.FAT12_SPECIAL_EOC:
                # Special EOC
                yield i
                return
            elif eoc_min <= self.fat[i] <= eoc_max:
                # End of cluster, end chain
                yield i
                return
            elif self.fat[i] == cluster_vals["BAD_CLUSTER"]:
                # Bad cluster, cannot follow chain, file broken!
                raise PyFATException("Bad cluster found in FAT cluster "
                                     "chain, cannot access file")
            elif self.fat[i] == cluster_vals["FREE_CLUSTER"]:
                # FREE_CLUSTER mark when following a chain is treated an error
                raise PyFATException("FREE_CLUSTER mark found in FAT cluster "
                                     "chain, cannot access file")
            else:
                raise PyFATException("Invalid or unknown FAT cluster "
                                     "entry found with value "
                                     "\'{}\'".format(hex(self.fat[i])))

            i = self.fat[i]

    @_init_check
    def close(self):
        """Close session and free up all handles."""
        self.__fp.close()
        self.initialized = False

    def __del__(self):
        """Try to close open handles."""
        try:
            self.close()
        except PyFATException:
            pass

    def __determine_fat_type(self) -> Union["PyFat.FAT_TYPE_FAT12",
                                            "PyFat.FAT_TYPE_FAT16",
                                            "PyFat.FAT_TYPE_FAT32"]:
        """Determine FAT type.

        An internal method to determine whether this volume is FAT12,
        FAT16 or FAT32.

        returns: `str`: Any of PyFat.FAT_TYPE_FAT12, PyFat.FAT_TYPE_FAT16
                 or PyFat.FAT_TYPE_FAT32
        """
        total_sectors = self._get_total_sectors()
        rsvd_sectors = self.bpb_header["BPB_RsvdSecCnt"]
        fat_sz = self.bpb_header["BPB_NumFATs"] * self._fat_size
        root_dir_sectors = self.root_dir_sectors
        data_sec = total_sectors - (rsvd_sectors + fat_sz + root_dir_sectors)
        count_of_clusters = data_sec // self.bpb_header["BPB_SecPerClus"]

        if count_of_clusters < 4085:
            msft_fat_type = self.FAT_TYPE_FAT12
        elif count_of_clusters < 65525:
            msft_fat_type = self.FAT_TYPE_FAT16
        else:
            msft_fat_type = self.FAT_TYPE_FAT32

        if self.bpb_header["BPB_FATSz16"] == 0:
            if self.bpb_header["BPB_FATSz32"] != 0:
                linux_fat_type = self.FAT_TYPE_FAT32
            else:
                linux_fat_type = msft_fat_type
        elif count_of_clusters >= 4085:
            linux_fat_type = self.FAT_TYPE_FAT16
        else:
            linux_fat_type = self.FAT_TYPE_FAT12

        if msft_fat_type != linux_fat_type:
            warnings.warn(f"Unable to reliably determine FAT type, "
                          f"guessing either FAT{msft_fat_type} or "
                          f"FAT{linux_fat_type}. Opting for "
                          f"FAT{linux_fat_type}.")
        return linux_fat_type

    def parse_header(self):
        """Parse BPB & FAT headers in opened file."""
        with self.__lock:
            self.__seek(0)
            boot_sector = self.__fp.read(512)

        self.bpb_header = BootSectorHeader()
        self.bpb_header.parse_header(boot_sector[:36])

        # Verify BPB headers
        self.__verify_bpb_header()

        # Parse FAT type specific header
        self.bpb_header = FAT12BootSectorHeader() \
            if self.bpb_header["BPB_FATSz16"] > 0 else FAT32BootSectorHeader()
        self.bpb_header.parse_header(boot_sector)

        # Determine FAT type
        self._fat_size = self._get_fat_size_count()
        self.fat_type = self.__determine_fat_type()

        # Calculate root directory sectors and starting point of root directory
        root_entries = self.bpb_header["BPB_RootEntCnt"]
        hdr_size = FATDirectoryEntry.FAT_DIRECTORY_HEADER_SIZE
        bytes_per_sec = self.bpb_header["BPB_BytsPerSec"]
        rsvd_secs = self.bpb_header["BPB_RsvdSecCnt"]
        num_fats = self.bpb_header["BPB_NumFATs"]

        self.root_dir_sectors = ((root_entries * hdr_size) +
                                 (bytes_per_sec - 1)) // bytes_per_sec
        self.root_dir_sector = rsvd_secs + (self._fat_size * num_fats)

        # Calculate first data sector
        self.first_data_sector = (rsvd_secs + (num_fats * self._fat_size) +
                                  self.root_dir_sectors)

        # Check signature
        with self.__lock:
            self.__seek(510)
            signature = struct.unpack("<H", self.__fp.read(2))[0]

        if signature != 0xAA55:
            raise PyFATException(f"Invalid signature: \'{hex(signature)}\'.")

        # Initialization finished
        self.initialized = True

    def __verify_bpb_header(self):
        """Verify BPB header for correctness."""
        if self.bpb_header["BS_jmpBoot"][0] == 0xEB:
            if self.bpb_header["BS_jmpBoot"][2] != 0x90:
                raise PyFATException("Boot code must end with 0x90")
        elif self.bpb_header["BS_jmpBoot"][0] == 0xE9:
            pass
        else:
            raise PyFATException("Boot code must start with 0xEB or "
                                 "0xE9. Is this a FAT partition?")

        #: 512,1024,2048,4096: As per fatgen103.doc
        byts_per_sec_range = [2**x for x in range(9, 13)]
        if self.bpb_header["BPB_BytsPerSec"] not in byts_per_sec_range:
            raise PyFATException(f"Expected one of {byts_per_sec_range} "
                                 f"bytes per sector, got: "
                                 f"\'{self.bpb_header['BPB_BytsPerSec']}\'.")

        #: 1,2,4,8,16,32,64,128: As per fatgen103.doc
        sec_per_clus_range = [2**x for x in range(8)]
        if self.bpb_header["BPB_SecPerClus"] not in sec_per_clus_range:
            raise PyFATException(f"Expected one of {sec_per_clus_range} "
                                 f"sectors per cluster, got: "
                                 f"\'{self.bpb_header['BPB_SecPerClus']}\'.")

        bytes_per_cluster = self.bpb_header["BPB_BytsPerSec"]
        bytes_per_cluster *= self.bpb_header["BPB_SecPerClus"]
        if bytes_per_cluster > 32768:
            warnings.warn("Bytes per cluster should not be more than 32K, "
                          "but got: {}K. Trying to continue "
                          "anyway.".format(bytes_per_cluster // 1024), Warning)

        if self.bpb_header["BPB_RsvdSecCnt"] == 0:
            raise PyFATException("Number of reserved sectors must not be 0")

        if self.bpb_header["BPB_Media"] not in [0xf0, 0xf8, 0xf9, 0xfa, 0xfb,
                                                0xfc, 0xfd, 0xfe, 0xff]:
            raise PyFATException("Invalid media type")

        if self.bpb_header["BPB_NumFATs"] < 1:
            raise PyFATException("At least one FAT expected, None found.")

        root_entry_count = self.bpb_header["BPB_RootEntCnt"] * 32
        root_entry_count %= self.bpb_header["BPB_BytsPerSec"]
        if self.bpb_header["BPB_RootEntCnt"] != 0 and root_entry_count != 0:
            raise PyFATException("Root entry count does not cleanly align with"
                                 " bytes per sector!")

        if self.bpb_header["BPB_TotSec16"] == 0 and \
                self.bpb_header["BPB_TotSec32"] == 0:
            raise PyFATException("16-Bit and 32-Bit total sector count "
                                 "value empty.")

    @staticmethod
    @contextmanager
    def open_fs(filename: str, offset: int = 0,
                encoding=FAT_OEM_ENCODING):
        """Context manager for direct use of PyFAT."""
        pf = PyFat(encoding=encoding, offset=offset)
        pf.open(filename)
        yield pf
        pf.close()



# -*- coding: utf-8 -*-

"""
Python FAT filesystem module with :doc:`PyFilesystem2 <pyfilesystem2:index>` \
compatibility.

pyfatfs allows interaction with FAT12/16/32 filesystems, either via
:doc:`PyFilesystem2 <pyfilesystem2:index>` for file-level abstraction
or direct interaction with the filesystem for low-level access.
"""

#: Specifies default ("OEM") encoding
from pyfatfs._exceptions import PyFATException

FAT_OEM_ENCODING = 'ibm437'
#: Specifies the long file name encoding, which is always UTF-16 (LE)
FAT_LFN_ENCODING = 'utf-16-le'


def _init_check(func):
    def _wrapper(*args, **kwargs):
        initialized = args[0].initialized

        if initialized is True:
            return func(*args, **kwargs)
        else:
            raise PyFATException("Class has not yet been fully initialized, "
                                 "please instantiate first.")

    return _wrapper



# -*- coding: utf-8 -*-

"""Implementation of `FatIO` for basic I/O."""
import errno
import io
import threading
from typing import Union, Optional

from pyfatfs import PyFATException

from pyfatfs.PyFat import PyFat


class DefaultModeReadOnly:
    def __init__(self):
        self.truncate = False
        self.appending = False
        self.reading = True
        self.writing = False


class FatIO(io.RawIOBase):
    """Wrap basic I/O operations for PyFat."""

    def __init__(self, fs: PyFat,
                 path: str,
                 mode = DefaultModeReadOnly()) -> None:
        """Wrap basic I/O operations for PyFat. **Currently read-only**.

        :param fs: `PyFat`: Instance of opened filesystem
        :param path: `str`: Path to file. If `mode` is *r*,
                            the file must exist.
        :param mode: `Mode`: Mode to open file in.
        """
        super(FatIO, self).__init__()
        self.mode = mode
        self.fs = fs
        self.name = str(path)
        self._lock = threading.Lock()

        self.dir_entry = self.fs.root_dir.get_entry(path)
        if self.dir_entry.is_directory() or self.dir_entry.is_special():
            raise IsADirectoryError(errno.EISDIR, path)
        elif self.dir_entry.is_volume_id():
            raise FileNotFoundError(errno.ENOENT, path)

        #: Position in bytes from beginning of file
        self.__bpos = 0
        #: Current cluster chain number
        self.__cpos = self.dir_entry.get_cluster()
        #: Current cluster chain index
        self.__cindex = 0
        #: Current cluster chain offset (in bytes)
        self.__coffpos = 0

        if self.mode.truncate:
            self.seek(0)
            self.truncate()
        if self.mode.appending:
            self.seek(0, 2)

    def __repr__(self) -> str:
        """Readable representation of class instance.

        ex: <FatFile fs=<PyFat object> path="/README.txt" mode="r">
        """
        return str(f'<{self.__class__.__name__} '
                   f'fs={self.fs} '
                   f'path="{self.name}" '
                   f'mode="{self.mode}">')

    def seek(self, offset: int, whence: int = 0) -> int:
        """Seek to a given offset in the file.

        :param offset: ``int``: offset in bytes in the file
        :param whence: ``int``: offset position:
                       - ``0``: absolute
                       - ``1``: relative to current position
                       - ``2``: relative to file end
        :returns: New position in bytes in the file
        """
        if whence == 1:
            offset += self.__bpos
        elif whence == 2:
            offset += self.dir_entry.filesize
        elif whence != 0:
            raise ValueError(f"Invalid whence {whence}, should be 0, 1 or 2")

        offset = min(offset, self.dir_entry.filesize)
        prev_index = self.__cindex

        self.__cindex = offset // self.fs.bytes_per_cluster
        self.__coffpos = offset % self.fs.bytes_per_cluster
        self.__bpos = offset

        if self.__bpos == self.dir_entry.filesize and \
                self.__bpos > 0 and self.__coffpos == 0:
            # We are currently at the end of the last cluster, there is no
            # next cluster so go back to the end of the previous cluster
            self.__coffpos = self.fs.bytes_per_cluster
            self.__cindex -= 1

        # If we go back, we have to start from the beginning of the file
        if self.__cindex < prev_index:
            self.__cpos = self.dir_entry.get_cluster()
            prev_index = 0

        if self.__cindex > prev_index:
            fp = self.fs.get_cluster_chain(self.__cpos)
            for _ in range(0, self.__cindex - prev_index + 1):
                self.__cpos = next(fp)

        return self.__bpos

    def seekable(self) -> bool:
        """FAT I/O driver is able to seek in files.

        :returns: `True`
        """
        return True

    def close(self) -> None:
        """Close open file handles assuming lock handle."""
        self.seek(0)
        if self.mode.writing:
            self.fs.flush_fat()
        super().close()

    def readable(self) -> bool:
        """Determine whether the file is readable."""
        return self.mode.reading

    def read(self, size: int = -1) -> Union[bytes, None]:
        """Read given bytes from file."""
        if not self.mode.reading:
            raise IOError("File not open for reading")

        with self._lock:
            # Set size boundary
            if size + self.__bpos > self.dir_entry.filesize or size < 0:
                size = self.dir_entry.filesize - self.__bpos

            if size == 0:
                return b""

            chunks = []
            read_bytes = 0
            cluster_offset = self.__coffpos
            for c in self.fs.get_cluster_chain(self.__cpos):
                chunk_size = self.fs.bytes_per_cluster - cluster_offset
                # Do not read past EOF
                if read_bytes + chunk_size > size:
                    chunk_size = size - read_bytes

                chunk = self.fs.read_cluster_contents(c)
                chunk = chunk[cluster_offset:][:chunk_size]
                cluster_offset = 0
                chunks.append(chunk)
                read_bytes += chunk_size
                if read_bytes == size:
                    break

            self.seek(read_bytes, 1)

            chunks = b"".join(chunks)
            if len(chunks) != size:
                raise RuntimeError("Read a different amount of data "
                                   "than was requested.")
            return chunks

    def readinto(self, __buffer: bytearray) -> Optional[int]:
        """Read data "directly" into bytearray."""
        data = self.read(len(__buffer))
        bytes_read = len(data)
        __buffer[:bytes_read] = data
        return bytes_read



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import string
import sys

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.compressions import (
    check_for_sequence,
    has_matching_alphabets,
    strip_suffix_from_archive,
    strip_suffix_from_compressed_file,
)
from ratarmountcore.utils import HEX, format_number

pytestmark = pytest.mark.order(0)


def test_strip_suffix_from_compressed_file():
    ssc = strip_suffix_from_compressed_file

    assert ssc('a.tar.bz2') == 'a.tar'
    assert ssc('a.tar.BZ2') == 'a.tar'
    assert ssc('a.tar.BZIP2') == 'a.tar'
    assert ssc('a.tar.gz') == 'a.tar'
    assert ssc('a.tar.gzip') == 'a.tar'
    assert ssc('a.tar.xz') == 'a.tar'
    assert ssc('a.tar.zst') == 'a.tar'
    assert ssc('a.tar') == 'a.tar'
    assert ssc('a.mp3') == 'a.mp3'


def test_strip_suffix_from_archive():
    sst = strip_suffix_from_archive

    assert sst('a.tar.bz2') == 'a'
    assert sst('a.tar.BZ2') == 'a'
    assert sst('a.tar.BZIP2') == 'a'
    assert sst('a.tar.gz') == 'a'
    assert sst('a.tar.gzip') == 'a'
    assert sst('a.tar.xz') == 'a'
    assert sst('a.tar.zst') == 'a'
    assert sst('a.tar') == 'a'
    assert sst('a.mp3') == 'a.mp3'

    assert sst('a.tbz2') == 'a'
    assert sst('a.TBZ2') == 'a'
    assert sst('a.tgz') == 'a'
    assert sst('a.txz') == 'a'
    assert sst('a.tzst') == 'a'


def test_has_matching_alphabets():
    matches = has_matching_alphabets

    assert matches('a', 'b')
    assert matches('0', '1')
    assert matches('0', 'a')  # because both might be hexadecimal
    assert matches('1a', 'b0')
    assert matches(HEX, HEX)
    assert matches(string.digits, string.digits)
    assert matches(string.ascii_lowercase, string.ascii_lowercase)
    assert matches(HEX, string.digits)
    assert not matches(string.ascii_lowercase, HEX)
    assert not matches(string.ascii_lowercase, string.digits)
    assert not matches('ag', 'b0')


def test_check_for_sequence():
    def to_alpha1(i):
        return format_number(i, string.ascii_lowercase, 1)

    assert check_for_sequence(['a'], to_alpha1) == ['a']
    assert check_for_sequence(['a', 'b'], to_alpha1) == ['a', 'b']
    assert check_for_sequence(['b', 'a'], to_alpha1) == ['a', 'b']
    assert check_for_sequence(['b', 'a', 'd'], to_alpha1) == ['a', 'b']
    assert check_for_sequence(['a', 'd'], to_alpha1) == ['a']
    assert check_for_sequence(['a', 'd', 'e'], to_alpha1) == ['a']
    assert not check_for_sequence(['aa'], to_alpha1)
    assert not check_for_sequence(['0'], to_alpha1)

    def to_alpha2(i):
        return format_number(i, string.ascii_lowercase, 2)

    assert check_for_sequence(['aa'], to_alpha2) == ['aa']
    assert check_for_sequence(['aa', 'ab'], to_alpha2) == ['aa', 'ab']
    assert check_for_sequence(['ab', 'aa'], to_alpha2) == ['aa', 'ab']
    assert check_for_sequence(['ab', 'aa', 'ad'], to_alpha2) == ['aa', 'ab']
    assert check_for_sequence(['aa', 'ad'], to_alpha2) == ['aa']
    assert check_for_sequence(['aa', 'ad', 'ae'], to_alpha2) == ['aa']
    assert not check_for_sequence(['aaa'], to_alpha2)
    assert not check_for_sequence(['0'], to_alpha2)
    assert not check_for_sequence(['00'], to_alpha2)

    def to_digit3(i):
        return format_number(i, string.digits, 3)

    assert check_for_sequence(['000'], to_digit3) == ['000']
    assert check_for_sequence(['000', '001'], to_digit3) == ['000', '001']
    assert check_for_sequence(['001', '000'], to_digit3) == ['000', '001']
    assert check_for_sequence(['001', '000', '003'], to_digit3) == ['000', '001']
    assert check_for_sequence(['000', '003'], to_digit3) == ['000']
    assert check_for_sequence(['000', '003', '004'], to_digit3) == ['000']
    assert not check_for_sequence(['0001'], to_digit3)



import builtins
import posixpath
from collections.abc import Iterable
from typing import IO, Any, Callable, Optional, Union

from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info
from ratarmountcore.utils import overrides


class RemovePrefixMountSource(MountSource):
    """
    MountSource for remove a prefix from all requested paths to effectively move the specified MountSource
    into a subfolder.
    """

    def __init__(self, path: str, mountSource: MountSource):
        # Beware, normpath leaves leading // but collapses /// and more repetitions to /!
        self.prefix = posixpath.normpath('/' + path.lstrip('/')).rstrip('/') + '/'
        self.mountSource = mountSource
        self._directory_info = create_root_file_info([])

    def _parse_path(
        self, path: str, queryMountSource: Callable[[str], Any], queryDir: Callable[[str], Any], defaultValue
    ):
        if not path.startswith('/'):
            path = '/' + path
        if not path.endswith('/'):
            path = path + '/'
        if path.startswith(self.prefix):
            return queryMountSource(path[len(self.prefix) - 1 : -1])
        if self.prefix.startswith(path):
            return queryDir(self.prefix.removeprefix(path).split('/', maxsplit=1)[0])
        return defaultValue

    # Methods with 'path' argument

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], dict[str, FileInfo]]]:
        return self._parse_path(path, self.mountSource.list, lambda name: {name: self._directory_info.clone()}, None)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], dict[str, int]]]:
        return self._parse_path(path, self.mountSource.list_mode, lambda name: {name: self._directory_info.mode}, None)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        return self._parse_path(path, self.mountSource.lookup, lambda _: self._directory_info.clone(), None)

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return self._parse_path(path, self.mountSource.versions, lambda _: 1, 0)

    @overrides(MountSource)
    def exists(self, path: str) -> bool:
        return self._parse_path(path, self.mountSource.exists, lambda _: True, False)

    @overrides(MountSource)
    def is_dir(self, path: str) -> bool:
        return self._parse_path(path, self.mountSource.is_dir, lambda _: True, False)

    # Methods with 'fileInfo' argument. Comparing against self._directory_info is not beautiful but should work
    # good enough. The alternative would be to add some token to FileInfo.userdata, but that seems too verbose.

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        if fileInfo == self._directory_info:
            raise FileNotFoundError("Cannot open directory!")
        return self.mountSource.open(fileInfo, buffering)

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        if fileInfo == self._directory_info:
            raise FileNotFoundError("Cannot read from directory!")
        return self.mountSource.read(fileInfo, size, offset)

    @overrides(MountSource)
    def get_mount_source(self, fileInfo: FileInfo):
        if fileInfo == self._directory_info:
            return super().get_mount_source(fileInfo)
        return self.mountSource.get_mount_source(fileInfo)

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> builtins.list[str]:
        if fileInfo == self._directory_info:
            return []
        return self.mountSource.list_xattr(fileInfo)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        if fileInfo == self._directory_info:
            return None
        return self.mountSource.get_xattr(fileInfo, key)

    # Methods that can simply be forwarded

    @overrides(MountSource)
    def statfs(self) -> dict[str, Any]:
        return self.mountSource.statfs()

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return self.mountSource.is_immutable()

    @overrides(MountSource)
    def __enter__(self):
        self.mountSource.__enter__()
        return self

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.mountSource.__exit__(exception_type, exception_value, exception_traceback)



# -*- coding: utf-8 -*-

"""FSInfo structure handling for FAT32 filesystems."""

import struct
from collections import OrderedDict


class FSInfo(OrderedDict):
    """Class for FAT32 FSInfo sector implementation."""

    #: Magic value at the start of an FSInfo struct
    LEAD_SIG_MAGIC = 0x41615252
    #: Magic value in the middle of an FSInfo struct
    STRUCT_SIG_MAGIC = 0x61417272
    #: Magic value at the end of an FSInfo struct
    TRAIL_SIG_MAGIC = 0xAA550000

    #: FSI header layout in struct formatted string
    HEADER_LAYOUT = "<L480xLLL12xL"
    #: FSI header fields
    HEADER_VARS = ["FSI_LeasSig",
                   "FSI_StrucSig",
                   "FSI_Free_Count",
                   "FSI_Nxt_Free",
                   "FSI_TrailSig"]

    def __init__(self, free_count=None, next_free=None):
        """Initialize an empty Boot sector header."""
        super().__init__()
        self.update({"FSI_LeasSig": FSInfo.LEAD_SIG_MAGIC,
                     "FSI_StrucSig": FSInfo.STRUCT_SIG_MAGIC,
                     "FSI_Free_Count": free_count,
                     "FSI_Nxt_Free": next_free,
                     "FSI_TrailSig": FSInfo.TRAIL_SIG_MAGIC})

    def __bytes__(self):
        """Serialize header data back to bytes."""
        return struct.pack(self.HEADER_LAYOUT, *self.values())

    def __len__(self):
        """Return struct size."""
        return struct.calcsize(self.HEADER_LAYOUT)

    def parse_header(self, data: bytes) -> "FSInfo":
        """Deserialize FSInfo binary data into FSInfo class instance.

        :param data: `bytes`: 512 bytes of binary data to be deserialized
        """
        if len(data) < len(self):
            raise ValueError(f"Invalid FSInfo sector data supplied "
                             f"for {type(self)}. len(data)={len(data)}, "
                             f"len(header)={len(self)}")

        fsinfo = struct.unpack(self.HEADER_LAYOUT, data[:len(self)])
        self.update(dict(zip(self.HEADER_VARS, fsinfo)))



import builtins
import posixpath
from collections.abc import Iterable
from typing import IO, Any, Optional, Union

from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info
from ratarmountcore.utils import RatarmountError, overrides


class SubvolumesMountSource(MountSource):
    def __init__(self, mountSources: dict[str, MountSource]) -> None:
        """
        mountSources : List of mount sources to mount as subfolders.
        """
        # Recursive dictionary representing folders. Keys are folder names, i.e., may not contain slashes.
        self.mountSources: dict[str, Any] = {}
        self.rootFileInfo = create_root_file_info(userdata=[None])

        for path, target in mountSources.items():
            self.mount(path, target)

    def _get_by_path(self, path: str, create: bool = False) -> Optional[Union[MountSource, dict[str, Any]]]:
        """Implements recursive lookup in self.mountSources."""
        folder = self.mountSources
        for part in path.strip('/').split('/'):
            if not part:
                continue
            if not isinstance(folder, dict):
                raise RatarmountError(f"Cannot return '{path}' because one of its parents is a mount point!")
            if part not in folder:
                if not create:
                    return None
                folder[part] = {}
            folder = folder[part]
        return folder

    def is_mountable(self, path: str) -> bool:
        # Ensuring a leading slash before calling normpath has the effect of eating all leading '/..'.
        parent, name = ('/' + posixpath.normpath('/' + path).strip('/')).rsplit('/', maxsplit=1)
        try:
            folder = self._get_by_path(parent, create=True)
        except RatarmountError:
            return False
        return bool(name) and folder is not None and isinstance(folder, dict) and (name not in folder)

    def mount(self, path: str, target: MountSource) -> bool:
        """Adds a mount source at the specified path. Must not overlap with existing mounts. Return true on success."""

        # Ensuring a leading slash before calling normpath has the effect of eating all leading '/..'.
        parent, name = ('/' + posixpath.normpath('/' + path).strip('/')).rsplit('/', maxsplit=1)
        if not name:
            raise RatarmountError("Mount points may not be empty!")

        try:
            folder = self._get_by_path(parent, create=True)
        except RatarmountError:
            return False

        if folder is None or not isinstance(folder, dict) or name in folder:
            raise RatarmountError("Mount point already exists!")
        folder[name] = target
        return True

    def unmount(self, path: str) -> Optional[MountSource]:
        """
        Removes a mounted path and returns the corresponding object.
        The caller can then call 'close' on it if necessary!
        """

        # Ensuring a leading / before calling normpath has the effect of eating all leading '/..'.
        path = posixpath.normpath('/' + path).strip('/')

        parents = [self.mountSources]
        names: list[str] = []
        for part in path.strip('/').split('/'):
            if not part:
                continue
            folder = parents[-1]
            names.append(part)
            if not isinstance(folder, dict) or part not in folder:
                return None
            parents.append(folder[part])

        # Happens if path is '/'.
        if not names:
            return None

        mountSource = parents.pop()
        if not isinstance(mountSource, MountSource):
            # Might happen when trying to unmount some parent path of an actual mount point.
            return None

        # Remove the mount source from its parent dictionary.
        parents.pop().pop(names.pop())
        # Remove empty parent folders from the recursive dictionary.
        for folder, name in zip(reversed(parents), reversed(names)):
            if folder[name]:
                break
            folder.pop(name)

        return mountSource

    def _find_mount_source(self, path: str) -> Optional[tuple[str, str, Union[MountSource, dict[str, Any]]]]:
        """
        Implements recursive lookup in self.mountSources.
        Returns (mount point path, path in mount source, mount source).
        If the path points to a parent mount folder, then will return (path, "", None)
        """

        folder = self.mountSources
        parts = path.strip('/').split('/')
        for i, part in enumerate(parts):
            if not part:
                continue
            if part not in folder:
                return None
            folder = folder[part]
            if isinstance(folder, MountSource):
                return ('/'.join(parts[: i + 1]), '/' + '/'.join(parts[i + 1 :]), folder)
        return (path, "", folder)

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return False  # Because we support mount and unmount!

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        result = self._find_mount_source(path)
        if result is None:
            return None

        subvolume, subpath, mounted = result
        if not isinstance(mounted, MountSource):
            return self.rootFileInfo.clone()

        fileInfo = mounted.lookup(subpath, fileVersion=fileVersion)
        if isinstance(fileInfo, FileInfo):
            fileInfo.userdata.append(subvolume)
            return fileInfo
        return None

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        result = self._find_mount_source(path)
        if result is None:
            return 0

        _subvolume, subpath, mounted = result
        if not isinstance(mounted, MountSource):
            return 1

        return mounted.versions(subpath)

    def _list(self, path: str, onlyMode: bool) -> Optional[Union[Iterable[str], dict[str, FileInfo]]]:
        result = self._find_mount_source(path)
        if result is None:
            return None

        _subvolume, subpath, mounted = result
        if not isinstance(mounted, MountSource):
            return {name: self.rootFileInfo.mode if onlyMode else self.rootFileInfo.clone() for name in mounted}

        return mounted.list_mode(subpath) if onlyMode else mounted.list(subpath)

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], dict[str, FileInfo]]]:
        return self._list(path, onlyMode=False)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], dict[str, int]]]:
        return self._list(path, onlyMode=True)

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        subvolume = fileInfo.userdata.pop()
        if subvolume is None:
            raise ValueError(f"Found subvolume is None for fileInfo: {fileInfo}")
        try:
            return self.mountSources[subvolume].open(fileInfo, buffering=buffering)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        subvolume = fileInfo.userdata.pop()
        if subvolume is None:
            raise ValueError(f"Found subvolume is None for fileInfo: {fileInfo}")
        try:
            return self.mountSources[subvolume].read(fileInfo, size, offset)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> builtins.list[str]:
        subvolume = fileInfo.userdata.pop()
        try:
            return [] if subvolume is None else self.mountSources[subvolume].list_xattr(fileInfo)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        subvolume = fileInfo.userdata.pop()
        try:
            return None if subvolume is None else self.mountSources[subvolume].get_xattr(fileInfo, key)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def get_mount_source(self, fileInfo: FileInfo) -> tuple[str, MountSource, FileInfo]:
        sourceFileInfo = fileInfo.clone()
        subvolume = sourceFileInfo.userdata.pop()

        if subvolume is None or subvolume not in self.mountSources:
            return '/', self, fileInfo
        mountSource = self.mountSources[subvolume]

        subpath, subMountSource, subFileInfo = mountSource.get_mount_source(sourceFileInfo)
        return posixpath.normpath(f"/{subvolume}/{subpath}"), subMountSource, subFileInfo

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        for mountSource in self.mountSources.values():
            mountSource.__exit__(exception_type, exception_value, exception_traceback)

    def join_threads(self):
        for mountSource in self.mountSources:
            if hasattr(mountSource, 'join_threads'):
                mountSource.join_threads()



import logging
import sys
import time
from typing import Any, Optional

try:
    import rich.progress
    from rich.logging import RichHandler
except ImportError:
    RichHandler = None  # type: ignore


def _logging_uses_rich() -> bool:
    return RichHandler is not None and any(isinstance(handler, RichHandler) for handler in logging.getLogger().handlers)


class ProgressBar:
    """Simple progress bar which keeps track of changes and prints the progress and a time estimate."""

    def __init__(self, maxValue: float):
        # Do not use thread_time here even if it is twice as fast because it does not count the time the thread
        # has been sleeping, which it does when it waits for I/O. This made the times incorrect for slow HDD accesses!
        self._get_time = time.time
        self.value = 0.0
        self.maxValue = maxValue
        self.lastUpdateTime = self._get_time()
        self.lastUpdateValue = 0.0
        self.updateInterval = 2.0  # seconds
        self.creationTime = self._get_time()
        self._richProgress: Optional[rich.progress.Progress] = None
        self._taskID: Optional[Any] = None

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback) -> None:
        if self._richProgress is None:
            return
        if self._taskID is not None:
            self._richProgress.update(self._taskID, completed=self.value)
            self._richProgress.refresh()
            # Calling remove_task would also remove the progress bar, which I don't want.
        self._richProgress.stop()
        self._richProgress.__exit__(exception_type, exception_value, exception_traceback)
        self._richProgress = None
        self._taskID = None

    def __del__(self) -> None:
        if self._richProgress:
            self._richProgress.stop()
            self._richProgress.__exit__(None, None, None)
            self._richProgress = None

    def start(self) -> None:
        if 'rich.progress' in sys.modules and self._richProgress is None and _logging_uses_rich():
            self._richProgress = rich.progress.Progress(
                rich.progress.TextColumn("[progress.description]{task.description}"),
                rich.progress.BarColumn(bar_width=None),
                rich.progress.TaskProgressColumn(),
                rich.progress.DownloadColumn(),
                rich.progress.TimeElapsedColumn(),
                rich.progress.TimeRemainingColumn(elapsed_when_finished=True),
                # Auto-refreshing and capturing is only pain, especially as I have already taken care to call
                # update in compute-intensive loops. There is no reason to fuck the console up only for a progress bar.
                auto_refresh=False,
                redirect_stdout=False,
                redirect_stderr=False,
            )

        if self._richProgress:
            self._richProgress.start()
            if self._taskID is None:
                self._taskID = self._richProgress.add_task("Processing", total=self.maxValue)
            self.updateInterval = 0.2

    def stop(self) -> None:
        if self._richProgress:
            self._richProgress.stop()

    def update(self, value: float) -> None:
        """Should be called whenever the monitored value changes. The progress bar is updated accordingly."""
        self.value = value
        if self.lastUpdateTime is not None and (self._get_time() - self.lastUpdateTime) < self.updateInterval:
            return

        if self._richProgress is None and 'rich.progress' in sys.modules:
            self.start()

        if self._richProgress and self._taskID is not None:
            self._richProgress.update(self._taskID, completed=value)
            self._richProgress.refresh()
        else:
            # Use whole interval since start to estimate time
            percent = value / self.maxValue if self.maxValue != 0 else 1.0
            totalTime = self._get_time() - self.creationTime
            eta1 = int(totalTime / percent - totalTime if percent != 0 else 0)
            # Use only a shorter window interval to estimate time.
            # Accounts better for higher speeds in beginning, e.g., caused by caching effects.
            # However, this estimate might vary a lot while the other one stabilizes after some time!
            if value == self.lastUpdateValue:
                eta2 = 99999 * 60
            else:
                eta2 = int(
                    (self._get_time() - self.lastUpdateTime) / (value - self.lastUpdateValue) * (self.maxValue - value)
                )
            print(
                f"Position {value} of {self.maxValue} ({percent * 100.0:.2f}%). "
                f"Remaining time: {eta2 // 60} min {eta2 % 60} s (current rate), "
                f"{eta1 // 60} min {eta1 % 60} s (average rate). "
                f"Spent time: {int(totalTime) // 60} min {int(totalTime) % 60} s",
                flush=True,
            )

        self.lastUpdateTime = self._get_time()
        self.lastUpdateValue = value



# libarchive.ffi produces many false positives on my local system and there seems to be nothing else works.
# pylint: disable=no-member

import contextlib
import ctypes
import io
import json
import logging
import os
import re
import stat
import sys
import tarfile
from collections.abc import Sequence
from pathlib import Path
from timeit import default_timer as timer
from typing import IO, Any, Callable, Optional, Union, cast

from ratarmountcore.compressions import COMPRESSION_BACKENDS
from ratarmountcore.formats import FILE_FORMATS
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.utils import InvalidIndexError, overrides

try:
    # Use the FFI directly because the higher-level interface does not work sufficiently for our use case.
    import libarchive.ffi as laffi
    from libarchive.exception import ArchiveError
except Exception:
    # I have encountered AttributeError, and on Windows also TypeError:
    #   File "C:\hostedtoolcache\windows\Python\3.11.9\x64\Lib\site-packages\libarchive\ffi.py", line 26, in <module>
    #     libarchive = ctypes.cdll.LoadLibrary(libarchive_path)
    #                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    #   File "C:\hostedtoolcache\windows\Python\3.11.9\x64\Lib\ctypes\__init__.py", line 454, in LoadLibrary
    #     return self._dlltype(name)
    #            ^^^^^^^^^^^^^^^^^^^
    #   File "C:\hostedtoolcache\windows\Python\3.11.9\x64\Lib\ctypes\__init__.py", line 366, in __init__
    #     if '/' in name or '\\' in name:
    #        ^^^^^^^^^^^
    #   TypeError: argument of type 'NoneType' is not iterable
    #
    # OSError can happen when dependencies are missing, e.g., libicuuc.so.74.
    pass

try:
    import py7zr  # pylint: disable=unused-import
except ImportError:
    py7zr = None  # type: ignore


logger = logging.getLogger(__name__)


class ArchiveEntry:
    def __init__(self, archive, entry_index: int, encoding: str = 'utf-8'):
        self._archive = archive  # Store for lifetime
        self._entry = laffi.entry_new()
        self.eof = laffi.read_next_header2(self._archive, self._entry) == laffi.ARCHIVE_EOF
        self.encoding = encoding
        self.entry_index = entry_index
        self._fileInfoRow: Optional[tuple] = None

    def __del__(self):
        laffi.entry_free(self._entry)

    def format_name(self):
        return laffi.format_name(self._archive)

    def get_time_by_name(self, name):
        if not getattr(laffi, f'entry_{name}_is_set')(self._entry):
            return None
        seconds = getattr(laffi, f'entry_{name}')(self._entry)
        nseconds = getattr(laffi, f'entry_{name}_nsec')(self._entry)
        return float(seconds) + float(nseconds) / 1e9 if nseconds else int(seconds)

    def get_time(self):
        result = self.get_time_by_name('mtime')
        if result is None:
            result = self.get_time_by_name('ctime')
        if result is None:
            result = self.get_time_by_name('birthtime')
        return result or 0

    def path(self) -> str:
        path = laffi.entry_pathname_w(self._entry)
        if not path:
            path = laffi.entry_pathname(self._entry)
            if path is not None:
                with contextlib.suppress(UnicodeError):
                    path = path.decode(self.encoding)
        return path

    def mode(self):
        return laffi.entry_mode(self._entry)

    def filetype(self):
        return laffi.entry_filetype(self._entry)

    def is_directory(self):
        return self.filetype() & 0o170000 == 0o040000

    def is_symbolic_link(self):
        return self.filetype() & 0o170000 == 0o120000

    def linkname(self):
        path = ""
        if self.is_symbolic_link():
            path = laffi.entry_symlink_w(self._entry)
            if not path:
                path = laffi.entry_symlink(self._entry)
        else:
            path = laffi.entry_hardlink_w(self._entry)
            if not path:
                path = laffi.entry_hardlink(self._entry)

        if isinstance(path, bytes):
            with contextlib.suppress(UnicodeError):
                path = path.decode(self.encoding)

        return path

    def convert_to_row(self, entryCount: int, transform: Callable[[str], str], path: Optional[str] = None) -> tuple:
        # The data logic may only be evaluated once because determining the size may require reading the whole file!
        if self._fileInfoRow is not None:
            return self._fileInfoRow

        if laffi.entry_size_is_set(self._entry):
            size = laffi.entry_size(self._entry)
        else:
            buffer = ctypes.create_string_buffer(1024 * 1024)
            size = 0
            while readSize := laffi.read_data(self._archive, buffer, len(buffer)):
                size += readSize

        mode = self.mode() & 0o777
        linkname = self.linkname()
        if linkname:
            mode = mode | stat.S_IFLNK
        else:
            linkname = ""
            mode = mode | (stat.S_IFDIR if self.is_directory() else stat.S_IFREG)

        if not path:
            path = self.path()
        path, name = SQLiteIndex.normpath(transform(path)).rsplit("/", 1)

        # Currently, this is unused. Only the headerOffset should be used.
        dataOffset = 0

        # The header offset needs to be set to some unique value in order to account for entries with identical paths.
        # We probably are not able to query the actual offset in the TAR via libarchive.
        # This makes the created index incompatible with the SQLiteIndexedTar backend.
        # For that reason more consistency checks were added as well as the backendName key in the index metadata.
        headerOffset = entryCount

        # fmt: off
        self._fileInfoRow = (
            path          ,  # 0  : path
            name          ,  # 1  : file name
            headerOffset  ,  # 2  : header offset
            dataOffset    ,  # 3  : data offset
            size          ,  # 4  : file size
            self.get_time(),  # 5  : modification time
            mode          ,  # 6  : file mode / permissions
            0             ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            linkname      ,  # 8  : linkname
            0             ,  # 9  : user ID
            0             ,  # 10 : group ID
            False         ,  # 11 : is TAR (unused?)
            False         ,  # 12 : is sparse
            False         ,  # 13 : is generated (parent folder)
            0             ,  # 14 : recursion depth
        )
        # fmt: on

        return self._fileInfoRow


class IterableArchive:
    """A class that opens an archive with libarchive and can iterate over the entries."""

    # Read formats list supported by python-libarchive-c:
    #   '7zip', 'all', 'ar', 'cab', 'cpio', 'empty', 'iso9660', 'lha', 'mtree',
    #   'rar', 'raw', 'tar', 'xar', 'zip', 'warc'
    # We especially do not want mtree because it has no magic bytes and basically matches anything with newlines,
    # including some test text files. 'raw', 'empty', and 'all' are special input formats.
    # It seems that 'raw' cannot be combined in an or-manner. It disables all archive format recognitions and only
    # applies the filters.
    # https://github.com/libarchive/libarchive/wiki/FormatRaw#caveat-dont-mix-_raw_-with-other-handlers
    # > aveat: Don't mix _raw_ with other handlers
    # > If you are using the raw handler, you should generally not enable any other handler
    # Note that the lz4 and zstd filters are not available in the manylinux2014 container (CentOS Linux
    # release 7.9.2009 (Core))
    ENABLED_FORMATS = ('7zip', 'ar', 'cab', 'cpio', 'iso9660', 'lha', 'rar', 'rar5', 'tar', 'xar', 'warc', 'zip')

    def __init__(
        self,
        file: Union[str, IO[bytes]],
        encoding='utf-8',
        passwords: Optional[Sequence[Union[str, bytes]]] = None,
        bufferSize=1024 * 1024,
    ):
        self.encoding = encoding
        self.bufferSize = bufferSize
        self._file = file
        self._entryIndex = 0  # A consecutive number to identify entries even if they have the same name.
        self._buffer = None
        self._readCallback = None
        self._seekCallback = None
        self._entry: Optional[ArchiveEntry] = None
        self._eof = False

        self._archive = laffi.read_new()
        try:
            self._set_passwords(passwords or [])
            self._try_to_open(allowArchives=True)
        except ArchiveError as exception:
            logger.info("Was not able to open %s as given archive. Try to undo compressions next.", self._file)
            logger.info("Exception: %s", exception, exc_info=logger.isEnabledFor(logging.DEBUG))
            try:
                laffi.read_free(self._archive)
                self._archive = laffi.read_new()

                self._set_passwords(passwords or [])
                self._try_to_open(allowArchives=False)
            except ArchiveError as exception2:
                raise exception2 from exception

    def format_name(self):
        return laffi.format_name(self._archive)

    def filter_names(self):
        allNames = [laffi.filter_name(self._archive, i) for i in range(laffi.filter_count(self._archive))]
        return [name for name in allNames if name != b'none']

    def _try_to_open(self, allowArchives: bool):
        laffi.get_read_filter_function('all')(self._archive)

        for formatToEnable in IterableArchive.ENABLED_FORMATS if allowArchives else ['raw']:
            try:
                laffi.get_read_format_function(formatToEnable)(self._archive)
            except ValueError as exception:
                # Ignore exceptions from formats that are not supported such as "rar5" in manylinux2014.
                logger.debug("Failed to enable format %s because of: %s", formatToEnable, exception, exc_info=True)

        if isinstance(self._file, str):
            laffi.read_open_filename_w(self._archive, self._file, os.stat(self._file).st_blksize)
        elif isinstance(self._file, Path):
            laffi.read_open_filename_w(self._archive, str(self._file), self._file.stat().st_blksize)
        elif isinstance(self._file, int):
            laffi.read_open_fd(self._archive, self._file, os.fstat(self._file).st_blksize)
        elif hasattr(self._file, 'readinto'):
            self._open_with_file_object()
        else:
            raise ValueError(
                "Libarchive backend currently only works with file path, descriptor, or a file object, "
                f"but got: {self._file}"
            )

        if not allowArchives and not self.filter_names():
            raise ArchiveError("When not looking for archives, there must be at least one filter!")

    def _set_passwords(self, passwords: Sequence[Union[str, bytes]]):
        try:
            for password in passwords:
                laffi.read_add_passphrase(self._archive, password.encode() if isinstance(password, str) else password)
        except AttributeError as exception:
            raise NotImplementedError(
                f"This libarchive library version ({laffi.version_number()}) at "
                f"{laffi.libarchive_path}) does not support encryption!"
            ) from exception

    def _open_with_file_object(self):
        if not hasattr(self._file, 'readinto'):
            return
        self._buffer = ctypes.create_string_buffer(self.bufferSize)

        def read_from_file_object(_archive, _context, pointerToBufferPointer):
            try:
                size = self._file.readinto(self._buffer)
            except (NotImplementedError, AttributeError):
                result = self._file.read(self.bufferSize)
                size = len(result)
                self._buffer = ctypes.c_char_p(result)
            pointerToBufferPointer = ctypes.cast(pointerToBufferPointer, ctypes.POINTER(ctypes.c_void_p))
            pointerToBufferPointer[0] = ctypes.cast(self._buffer, ctypes.c_void_p)
            return size

        if hasattr(self._file, 'seekable') and self._file.seekable() and hasattr(self._file, 'tell'):

            def seek_file_object(_archive, _context, offset, whence):
                self._file.seek(offset, whence)
                return self._file.tell()

            self._file.seek(0)
            self._seekCallback = laffi.SEEK_CALLBACK(seek_file_object)
            # Needs to be assigned to self, or else, ctypes lifetime issues result in a segfault!
            laffi.read_set_seek_callback(self._archive, self._seekCallback)

        # Needs to be assigned to self, or else, ctypes lifetime issues result in a segfault!
        self._readCallback = laffi.READ_CALLBACK(read_from_file_object)
        laffi.read_open(self._archive, None, laffi.NO_OPEN_CB, self._readCallback, laffi.NO_CLOSE_CB)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

    def __del__(self):
        laffi.read_free(self._archive)

    def next_entry(self) -> Optional[ArchiveEntry]:
        if self._eof:
            return None

        self._entry = ArchiveEntry(self._archive, entry_index=self._entryIndex, encoding=self.encoding)
        if self._entryIndex == 0 and logger.isEnabledFor(logging.INFO):
            format_name = laffi.format_name(self._archive)
            if isinstance(format_name, bytes):
                format_name = format_name.decode()
            # We need to try and read the first entry before format_name returns anything other than 'none'.
            logger.info(
                "Successfully opened type '%s' with libarchive. Using filters: %s", format_name, self.filter_names()
            )

        self._entryIndex += 1
        if self._entry.eof:
            self._eof = True
            return None

        return self._entry

    def entry_index(self):
        return self._entryIndex

    def read_data(self, buffer, size):
        return laffi.read_data(self._archive, buffer, size)


class IterableArchiveCache:
    def __init__(self, cacheSize: int = 5):
        self.cacheSize = cacheSize
        self._cache: list[IterableArchive] = []

    def insert(self, archive: IterableArchive):
        self._cache.append(archive)
        while len(self._cache) > self.cacheSize:
            self._cache.pop(0)

    def take(self, entry_index: int) -> Optional[IterableArchive]:
        qualified = [i for i in range(len(self._cache)) if self._cache[i].entry_index() < entry_index]
        if not qualified:
            return None
        # Find the archive with the closest (highest) index not larger or equal than the requested one
        result = max(qualified, key=lambda i: self._cache[i].entry_index())
        return self._cache.pop(result)


class LibarchiveFile(io.RawIOBase):
    def __init__(
        self,
        file,
        entry_index,
        fileSize,
        passwords: Optional[Sequence[str]] = None,
        archiveCache: Optional[IterableArchiveCache] = None,
    ):
        io.RawIOBase.__init__(self)
        self.file = file
        self.fileSize = fileSize
        self.entry_index = entry_index
        self.passwords = passwords
        self._archiveCache = archiveCache
        self._archive = None
        self._entry = None
        self._bufferSizeMax = 1024 * 1024
        self._buffer = None
        self._bufferOffset = 0  # Offset in self.entry from which the buffer was read
        self._bufferIO = None

        self._open()

    def _open(self):
        self.close()

        self._archive = self._archiveCache.take(self.entry_index)
        if self._archive is None:
            self._archive = IterableArchive(self.file, passwords=self.passwords)

        while True:
            self._entry = self._archive.next_entry()
            if self._entry is None:
                break
            if self._entry.entry_index == self.entry_index:
                self._refill_buffer()
                break

        if self._entry is None and self._bufferIO is None:
            raise ValueError(f"Failed to find archive entry {self.entry_index}.")

    def _refill_buffer(self):
        bufferedSize = len(self._buffer) if self._buffer else 0
        if self._bufferOffset + bufferedSize >= self.fileSize:
            return

        self._bufferOffset += bufferedSize
        sizeToRead = min(self._bufferSizeMax, max(0, self.fileSize - self._bufferOffset))

        # Reallocate buffer if we need a different size. This is necessary because I don't see an API to
        # specify a size to io,BytesIO to make it work on a subset of the buffer without copying.
        if sizeToRead >= self._bufferSizeMax:
            if self._buffer is None or len(self._buffer) < self._bufferSizeMax:
                self._buffer = ctypes.create_string_buffer(self._bufferSizeMax)
        elif self._buffer is None or len(self._buffer) != sizeToRead:
            self._buffer = ctypes.create_string_buffer(sizeToRead)

        readSize = self._archive.read_data(self._buffer, sizeToRead)
        if readSize != sizeToRead:
            raise RuntimeError(f"Read {readSize} bytes but expected {self.fileSize} for entry {self.entry_index}!")
        self._bufferIO = io.BytesIO(self._buffer)

    @overrides(io.RawIOBase)
    def close(self) -> None:
        if self._archiveCache and self._archive:
            self._archiveCache.insert(self._archive)
        if self._bufferIO is not None:
            self._bufferIO.close()
            self._bufferIO = None
            self._bufferOffset = 0
        self._buffer = None
        self._entry = None
        self._archive = None

    @overrides(io.RawIOBase)
    def fileno(self) -> int:
        # This is a virtual Python level file object and therefore does not have a valid OS file descriptor!
        raise io.UnsupportedOperation

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return self._bufferIO is not None and self._bufferIO.readable()

    @overrides(io.RawIOBase)
    def writable(self) -> bool:
        return False

    def read1(self, size: int = -1) -> bytes:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be read from!")
        result = self._bufferIO.read(size)
        if result:
            return result
        self._refill_buffer()
        return self._bufferIO.read(size)

    @overrides(io.RawIOBase)
    def read(self, size: int = -1) -> bytes:
        result = bytearray()
        while size < 0 or len(result) < size:
            read_data = self.read1(size if size < 0 else size - len(result))
            if not read_data:
                break
            result.extend(read_data)
        return bytes(result)

    def _skip(self, size: int) -> None:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be read from!")
        while size > 0:
            data = self.read1(size)
            if not data:
                break
            size -= len(data)

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be seeked!")

        here = self.tell()
        if whence == io.SEEK_CUR:
            offset += here
        elif whence == io.SEEK_END:
            offset += self.fileSize

        if offset >= here:
            self._skip(offset - here)
            return self.tell()

        bufferedSize = 0
        if self._buffer:
            bufferedSize = len(self._buffer)
        if offset >= self._bufferOffset and offset < self._bufferOffset + bufferedSize:
            self._bufferIO.seek(offset - self._bufferOffset)
        else:
            self._open()
            self._skip(max(0, offset))
        return self.tell()

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        result = self._bufferOffset
        if self._bufferIO:
            result += self._bufferIO.tell()
        return result


# The implementation is similar to ZipMountSource and SQLiteIndexedTarUserData.
class LibarchiveMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes]],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[Sequence[str]]   = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        transform              : Optional[tuple[str, str]] = None,
        indexMinimumFileCount  : int                       = 0,
        tarFileName            : Optional[str]             = None,
        **options
    ) -> None:
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.fileOrPath             = fileOrPath
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.options                = options
        self.transformPattern       = transform
        self.passwords              = options.get("passwords", [])
        self.tarFileName            = tarFileName
        self._archiveCache          = IterableArchiveCache()
        # fmt: on

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        # Determine an archive file name to show for debug output and as file name inside the mount point for
        # simple non-TAR gzip/bzip2 stream-compressed files.
        if tarFileName:
            self.tarFileName = tarFileName
        elif isinstance(fileOrPath, str):
            self.tarFileName = fileOrPath
        else:
            self.tarFileName = '<file object>'

        # Force indexes in memory because:
        #  - They are incompatible with SQLiteIndexedTar and relevant consistency checks do not exist in older versions.
        #    Older versions would simply show the wrong folder hierarchy and return Input/Output error on file access.
        #  - Seeking to a file takes on average half as time much as creating the index. I.e., the overhead for
        #    creating the index feels relatively insignificant assuming that more than 2 files are accessed.
        indexFilePath = ':memory:'
        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='LibarchiveMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        isFileObject = False  # Not supported yet

        if self.index.index_is_loaded():
            metadata = dict(self.index.get_connection().execute('SELECT * FROM metadata;'))
            if 'backend' not in metadata or metadata['backend'] != 'libarchive':
                self.__exit__(None, None, None)
                raise InvalidIndexError("The found index was not created by the libarchive backend.")

            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _create_index(self) -> None:
        if logger.isEnabledFor(logging.WARNING):
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()

        triedToOpen = False
        fileInfos = []
        gotAnyEntry = False
        with IterableArchive(self.fileOrPath, passwords=self.passwords) as archive:
            while True:
                entry = archive.next_entry()
                if entry is None:
                    break

                entryPath = None
                if entry.entry_index == 0 and self.tarFileName and entry.format_name() == b'raw':
                    libarchiveSuffixes = [
                        s for fid in COMPRESSION_BACKENDS['libarchive'].formats for s in FILE_FORMATS[fid].extensions
                    ]
                    fname = os.path.basename(self.tarFileName)
                    for suffix in ['gz', 'bz2', 'bzip2', 'gzip', 'xz', 'zst', 'zstd', *libarchiveSuffixes]:
                        suffix = '.' + suffix
                        if fname.lower().endswith(suffix.lower()) and len(fname) > len(suffix):
                            fname = fname[: -len(suffix)]
                            break
                    entryPath = fname

                gotAnyEntry = True
                fileInfos.append(entry.convert_to_row(entry.entry_index, self.transform, path=entryPath))
                # Contains file info SQLite row tuples! 4 -> size, 6 -> mode
                if not triedToOpen and not stat.S_ISDIR(fileInfos[-1][6]) and fileInfos[-1][4] > 0:
                    bufferSize = 1
                    buffer = ctypes.create_string_buffer(bufferSize)
                    try:
                        archive.read_data(buffer, 1)
                    except ArchiveError as exception:
                        # Very special case to delegate to py7zr somewhat smartly for encrypted 7z archives.
                        if entry.format_name() == b'7-Zip' and self.passwords and "py7zr" in sys.modules:
                            raise exception
                        if 'encrypt' in str(exception).lower():
                            logger.warning("The file contents are encrypted but not the file hierarchy!")
                            logger.warning("Specify a password with --password to also view file contents!")
                    triedToOpen = True

                if len(fileInfos) > 1000:
                    self.index.set_file_infos(fileInfos)
                    fileInfos = []

            if fileInfos:
                self.index.set_file_infos(fileInfos)
            elif not gotAnyEntry and archive.format_name() == b'tar':
                raise ArchiveError("Supposedly detected a TAR with no entries in it. Rejecting it as unknown format!")

        self.index.finalize()

        if logger.isEnabledFor(logging.WARNING):
            print(f"Creating offset dictionary for {self.archiveFilePath} took {timer() - t0:.2f}s")

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding', 'transformPattern']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)
        self.index.store_metadata_key_value('backend', 'libarchive')

    def __del__(self):
        # TODO check that all objects are really closed to avoid memory leaks
        pass

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        assert fileInfo.userdata
        tarFileInfo = fileInfo.userdata[-1]
        assert isinstance(tarFileInfo, SQLiteIndexedTarUserData)
        return cast(
            IO[bytes],
            LibarchiveFile(
                self.fileOrPath,
                tarFileInfo.offsetheader,
                fileSize=fileInfo.size,
                passwords=self.passwords,
                archiveCache=self._archiveCache,
            ),
        )

    def _check_metadata(self, metadata: dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(
                json.loads(metadata['arguments']), self, argumentsToCheck=['encoding', 'transformPattern']
            )



# -*- coding: utf-8 -*-

"""Boot sector header specific implementation for FAT12/16 and FAT32."""

import struct
from collections import OrderedDict


class BootSectorHeader(OrderedDict):
    """Base/Interface class for Boot sector header implementation."""

    #: BPB header layout in struct formatted string
    HEADER_LAYOUT = "<3s8sHBHBHHBHHHLL"
    #: BPB header fields when extracted with bpb_header_layout
    HEADER_VARS = ["BS_jmpBoot", "BS_OEMName", "BPB_BytsPerSec",
                   "BPB_SecPerClus", "BPB_RsvdSecCnt", "BPB_NumFATs",
                   "BPB_RootEntCnt", "BPB_TotSec16", "BPB_Media",
                   "BPB_FATSz16", "BPB_SecPerTrk", "BPB_NumHeads",
                   "BPB_HiddSec", "BPB_TotSec32"]

    def __init__(self):
        """Initialize an empty Boot sector header."""
        super().__init__()
        self.update(dict.fromkeys(self.HEADER_VARS))

    def __bytes__(self):
        """Serialize header data back to bytes."""
        return struct.pack(self.HEADER_LAYOUT, *self.values())

    def __len__(self):
        """Return struct size."""
        return struct.calcsize(self.HEADER_LAYOUT)

    def parse_header(self, data: bytes):
        """Parse header data from bytes.

        :param data: `bytes`: Raw header data from disk
        """
        if len(data) < len(self):
            raise ValueError(f"Invalid Boot sector header data supplied "
                             f"for {type(self)}. len(data)={len(data)}, "
                             f"len(header)={len(self)}")

        header = struct.unpack(self.HEADER_LAYOUT, data[:len(self)])
        self.update(dict(zip(self.HEADER_VARS, header)))


class FAT12BootSectorHeader(BootSectorHeader):
    """FAT12/16-specific header implementation."""

    #: FAT12/16 header layout in struct formatted string
    HEADER_LAYOUT = BootSectorHeader.HEADER_LAYOUT + "BBBL11s8s"
    #: FAT12/16 header fields when extracted with fat12_header_layout
    HEADER_VARS = BootSectorHeader.HEADER_VARS + \
        ["BS_DrvNum", "BS_Reserved1", "BS_BootSig",
         "BS_VolID", "BS_VolLab", "BS_FilSysType"]


class FAT32BootSectorHeader(BootSectorHeader):
    """FAT32-specific header implementation."""

    #: FAT32 header layout in struct formatted string
    HEADER_LAYOUT = BootSectorHeader.HEADER_LAYOUT + "LHHLHH12sBBBL11s8s"
    #: FAT32 header fields when extracted with fat32_header_layout
    HEADER_VARS = BootSectorHeader.HEADER_VARS + \
        ["BPB_FATSz32", "BPB_ExtFlags", "BPB_FSVer", "BPB_RootClus",
         "BPB_FSInfo", "BPB_BkBootSec", "BPB_Reserved", "BS_DrvNum",
         "BS_Reserved1", "BS_BootSig", "BS_VolID", "BS_VolLab",
         "BS_FilSysType"]
